@@ -14,6 +14,7 @@ __all__ = [
     'SOSPolynomialTransform',
     'FreeFormJacobianTransform',
     'AutoregressiveTransform',
+    'LULinearTransform',
     'PermutationTransform',
 ]
 
@@ -107,6 +108,17 @@ class ComposedTransform(Transform):
             x = t(x)
         return x
 
+    @property
+    def inv(self) -> Transform:
+        new = self.__new__(ComposedTransform)
+        new.transforms = [t.inv for t in reversed(self.transforms)]
+        new.domain_dim = self.codomain_dim
+        new.codomain_dim = self.domain_dim
+
+        Transform.__init__(new)
+
+        return new
+
     def _inverse(self, y: Tensor) -> Tensor:
         for t in reversed(self.transforms):
             y = t.inv(y)
@@ -585,7 +597,7 @@ class FreeFormJacobianTransform(Transform):
     The transformation is the integration of a system of first-order ordinary
     differential equations
 
-    .. math:: x(T) = \int_0^T f_\phi(t, x(t)) ~ dt .
+    .. math:: x(T) = x(0) + \int_0^T f_\phi(t, x(t)) ~ dt .
 
     References:
         | FFJORD: Free-form Continuous Dynamics for Scalable Reversible Generative Models (Grathwohl et al., 2018)
@@ -623,6 +635,20 @@ class FreeFormJacobianTransform(Transform):
     def _call(self, x: Tensor) -> Tensor:
         return odeint(self.f, x, self.t0, self.t1, self.phi)
 
+    @property
+    def inv(self) -> Transform:
+        new = self.__new__(FreeFormJacobianTransform)
+        new.f = self.f
+        new.t0 = self.t1
+        new.t1 = self.t0
+        new.phi = self.phi
+        new.exact = self.exact
+        new.trace_scale = self.trace_scale
+
+        Transform.__init__(new)
+
+        return new
+
     def _inverse(self, y: Tensor) -> Tensor:
         return odeint(self.f, y, self.t1, self.t0, self.phi)
 
@@ -634,28 +660,22 @@ class FreeFormJacobianTransform(Transform):
         if self.exact:
             I = torch.eye(x.shape[-1], dtype=x.dtype, device=x.device)
             I = I.expand(*x.shape, x.shape[-1]).movedim(-1, 0)
-
-            def f_aug(t: Tensor, x: Tensor, ladj: Tensor) -> Tensor:
-                with torch.enable_grad():
-                    x = x.requires_grad_().expand(I.shape)
-                    dx = self.f(t, x)
-
-                jacobian = torch.autograd.grad(dx, x, I, create_graph=True)[0]
-                trace = torch.einsum('i...i', jacobian)
-
-                return dx[0], trace * self.trace_scale
         else:
             eps = torch.randn_like(x)
 
-            def f_aug(t: Tensor, x: Tensor, ladj: Tensor) -> Tensor:
-                with torch.enable_grad():
-                    x = x.requires_grad_()
-                    dx = self.f(t, x)
+        def f_aug(t: Tensor, x: Tensor, ladj: Tensor) -> Tensor:
+            with torch.enable_grad():
+                x = x.requires_grad_()
+                dx = self.f(t, x)
 
+            if self.exact:
+                jacobian = torch.autograd.grad(dx, x, I, create_graph=True, is_grads_batched=True)[0]
+                trace = torch.einsum('i...i', jacobian)
+            else:
                 epsjp = torch.autograd.grad(dx, x, eps, create_graph=True)[0]
                 trace = (epsjp * eps).sum(dim=-1)
 
-                return dx, trace * self.trace_scale
+            return dx, trace * self.trace_scale
 
         ladj = torch.zeros_like(x[..., 0])
         y, ladj = odeint(f_aug, (x, ladj), self.t0, self.t1, self.phi)
@@ -706,6 +726,46 @@ class AutoregressiveTransform(Transform):
         return y, ladj.sum(dim=-1)
 
 
+class LULinearTransform(Transform):
+    r"""Creates a transformation :math:`f(x) = L U x`.
+
+    Arguments:
+        LU: A matrix whose lower and upper triangular parts are the non-zero elements
+            of :math:`L` and :math:`U`, with shape :math:`(*, D, D)`.
+    """
+
+    domain = constraints.real_vector
+    codomain = constraints.real_vector
+    bijective = True
+
+    def __init__(self, LU: Tensor, **kwargs):
+        super().__init__(**kwargs)
+
+        I = torch.eye(LU.shape[-1], dtype=LU.dtype, device=LU.device)
+
+        self.L = torch.tril(LU, diagonal=-1) + I
+        self.U = torch.triu(LU, diagonal=+1) + I
+
+    def _call(self, x: Tensor) -> Tensor:
+        return (self.L @ self.U @ x.unsqueeze(-1)).squeeze(-1)
+
+    def _inverse(self, y: Tensor) -> Tensor:
+        return torch.linalg.solve_triangular(
+            self.U,
+            torch.linalg.solve_triangular(
+                self.L,
+                y.unsqueeze(-1),
+                upper=False,
+                unitriangular=True,
+            ),
+            upper=True,
+            unitriangular=True,
+        ).squeeze(-1)
+
+    def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
+        return x.new_zeros(x.shape[:-1])
+
+
 class PermutationTransform(Transform):
     r"""Creates a transformation that permutes the elements.
 
